import {NS} from "@ns";

/**
 * tick in scheduler
 */
export interface Tickable {
    /**
     * tick in scheduler
     */
    tick(): void;
    /**
     * no work to run in this tick.
     */
    empty(): boolean;
}

/**
 * worker in scheduler
 */
export interface Work {
    work: (() => Promise<void>) | (()=>void);
}

/**
 * set of work to be executed at the same time
 */
export class WorkSet implements Work{
    works: Work[];
    constructor(){
        this.works = [];
    }
    add(work: Work): void{
        this.works.push(work);
    }
    remove(work: Work): void{
        this.works = this.works.filter(x=>x !== work);
    }
    concat(other: WorkSet): void{
        this.works = this.works.concat(other.works);
    }
    async work(): Promise<void>{
        for(const work of this.works){
            await work.work();
        }
    }
}

export function mergeWork(...args: Work[]): WorkSet{
    const workSet = new WorkSet();
    for(const work of args){
        if(work instanceof WorkSet){
            workSet.concat(work);
        }
        else {
            workSet.add(work);
        }
    }
    return workSet;
}

class PidTerminateExecutor implements Tickable{
    ns: NS;
    lastCheck: number;
    pidSet: Map<number, Work>;

    constructor(ns: NS){
        this.ns = ns;
        this.lastCheck = Date.now();
        this.pidSet = new Map();
    }
    watchPid(pid: number, work: Work): void{
        if(this.pidSet.has(pid)){
            throw new Error(`pid ${pid} is already watched`);
        }
        this.pidSet.set(pid, work);
    }
    
    empty(): boolean{
        return this.pidSet.size === 0;
    }

    async tick(): Promise<void>{
        const current =  Date.now();
        const deadScripts = this.ns.getRecentScripts().filter(x=>x.timeOfDeath.getTime() >= this.lastCheck);
        for(const script of deadScripts){
            const target = this.pidSet.get(script.pid);
            if(target){
                await target.work();
            }
            this.pidSet.delete(script.pid);
        }
        this.lastCheck = current;
    }
}

function normalizeTime(time: number, timeResolution: number): number{
    return Math.floor(time / timeResolution) * timeResolution;
}

class TimerExecutor implements Tickable{
    ns: NS;
    timerResolution: number;
    timerStart: number;
    workBucket: Map<number, WorkSet>;
    constructor(ns: NS, timerResolution = 250){
        this.ns = ns;
        this.workBucket = new Map;
        this.timerResolution = timerResolution;
        this.timerStart = performance.now();
    }

    empty(): boolean{
        return this.workBucket.size === 0;
    }

    setTimeout(work: Work, delay: number): void {
        const time = performance.now() + delay - this.timerStart;
        const normalizedTime = normalizeTime(time, this.timerResolution);
        const workSet = this.workBucket.get(normalizedTime);
        if(workSet){
            workSet.add(work);
        }
        else{
            this.workBucket.set(normalizedTime, mergeWork(work));
        }
    }
    async tick(): Promise<void> {
        const current = performance.now();
        const normalizedCurrent = normalizeTime(current-this.timerStart, this.timerResolution);
        const workSet = this.workBucket.get(normalizedCurrent);
        if(workSet){
            await workSet.work();
            this.workBucket.delete(normalizedCurrent);
        }
    }
    async sleepForTick(): Promise<void>{
        const current = performance.now();

        const normalizedCurrent = normalizeTime(current-this.timerStart, this.timerResolution);
        const next = normalizedCurrent + this.timerResolution;

        const nextTime = this.timerStart + next;
        const sleepTime = nextTime - current;
        //this.ns.print(`Sleeping for ${sleepTime}ms`);
        await this.ns.sleep(sleepTime);
    }
}

export class Scheduler{
    ns: NS;
    timeResolution: number;
    pidWatcher: PidTerminateExecutor;
    timerWatcher: TimerExecutor;
    constructor(ns: NS, timeResolution = 250){
        this.ns = ns;
        this.timeResolution = timeResolution;
        this.pidWatcher = new PidTerminateExecutor(ns);
        this.timerWatcher = new TimerExecutor(ns, timeResolution);
    }

    empty(): boolean{
        return this.pidWatcher.empty() && this.timerWatcher.empty();
    }

    setTimeout(work: Work, delay: number): void {
        this.timerWatcher.setTimeout(work, delay);
    }

    sleepInWork(delay: number): Promise<void>{
        return new Promise<void>(resolve=>{
            this.setTimeout({work(){
                resolve();
            } }, delay);
        });
    }
    onPidTerminate(pid: number, work: Work): void{
        this.pidWatcher.watchPid(pid, work);
    }

    async tick(): Promise<void>{
        await this.pidWatcher.tick();
        await this.timerWatcher.tick();
        await this.timerWatcher.sleepForTick();
    }
}